package cs108;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UncheckedIOException;
import java.util.Set;
import java.util.TreeSet;
import java.util.zip.GZIPInputStream;

public final class GameBoyCheater {
    static byte[] readMemDump(String fileName) {
        try (InputStream s = new GZIPInputStream(new FileInputStream(fileName))) {
            byte[] memDump = new byte[0x1_0000];
            int totalRead = 0;
            while (totalRead != memDump.length) {
                int read = s.read(memDump, totalRead, memDump.length - totalRead);
                assert read > 0;
                totalRead += read;
            }
            return memDump;
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    static Set<Integer> addressesContaining(byte[] dump, int v) {
        Set<Integer> s = new TreeSet<>();
        for (int a = 0; a < dump.length; ++a) {
            if (Byte.toUnsignedInt(dump[a]) == v)
                s.add(a);
        }
        return s;
    }

    static Set<Integer> addressesContaining2(byte[] dump, byte[] bytes) {
        Set<Integer> s = new TreeSet<>();
        for (int a = 0; a <= dump.length - bytes.length; ++a) {
            int equalBytes = 0;
            for (int i = 0; i < bytes.length && dump[a + i] == bytes[i]; ++i)
                equalBytes += 1;
            if (equalBytes == bytes.length)
                s.add(a);
        }
        return s;
    }

    // Variant using a named loop
    static Set<Integer> addressesContaining(byte[] dump, byte[] bytes) {
        Set<Integer> s = new TreeSet<>();
        outer:
            for (int a = 0; a <= dump.length - bytes.length; ++a) {
                for (int i = 0; i < bytes.length; ++i) {
                    if (dump[a + i] != bytes[i])
                        continue outer;
                }
                s.add(a);
            }
        return s;
    }

    public static void main(String[] args) {
        System.out.println("\n---------- Super Mario Land 2");
        superMarioLand2();
        System.out.println("\n---------- Tetris");
        tetris();
    }

    private static void superMarioLand2() {
        Set<Integer> s3 = addressesContaining(readMemDump("sml_3.bin.gz"), 3);
        Set<Integer> s4 = addressesContaining(readMemDump("sml_4.bin.gz"), 4);
        Set<Integer> s5 = addressesContaining(readMemDump("sml_5.bin.gz"), 5);
        System.out.printf("3 vies : %d candidats%n", s3.size());
        System.out.printf("4 vies : %d candidats%n", s4.size());
        System.out.printf("5 vies : %d candidats%n", s5.size());
    
        s5.retainAll(s4);
        s5.retainAll(s3);
    
        System.out.println("Adresses :");
        for (int a: s5)
            System.out.printf("  %04X%n", a);
    }

    private static void tetris() {
        Set<Integer> s1 = addressesContaining(readMemDump("tetris_11.bin.gz"), 0x11);
        Set<Integer> s2 = addressesContaining(readMemDump("tetris_799.bin.gz"), new byte[] {(byte) 0x99, 0x07});
        Set<Integer> s3 = addressesContaining(readMemDump("tetris_3979.bin.gz"), new byte[] {(byte) 0x79, 0x39});
        System.out.printf("score   11 : %d candidats%n", s1.size());
        System.out.printf("score  799 : %d candidats%n", s2.size());
        System.out.printf("score 3979 : %d candidats%n", s3.size());

        s1.retainAll(s2);
        s1.retainAll(s3);

        System.out.println("Adresses :");
        for (int a: s1)
            System.out.printf("  %04X%n", a);
    }
}
